#################################################################################
# Created       : Martin Foster
# Modified      : 02-Apr-2007
#################################################################################
#
# Param - Modules part of Ethereal Realms, designed to handle parameters
#         diffrently
# Copyright (C) 2000-2007  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 4-3526 Wolfe Cres
#                 Halifax, NS
#                 B3L 3S2
#
#################################################################################

package Ethereal::Param;					# Pacakge name

use DB_File;							# BerkeleyDB handle
use Storable qw(nfreeze thaw);					# Storable activity
use MIME::Base64;						# Shorter encoding
use strict;							# Strict variable enforcement


#################################################################################
# Sub-Routines
#################################################################################

#####################
# Constructor
#
# Allows to class to be exported and thus spawn objects as a result

sub new
{
	####################
	# Data members
	my $class    = shift;					# Class itself
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle

	my $error;						# Error code
	my $self;						# Self referncing link


	####################
	# Program area

	# Define parameters
	$self = {
	 XCGI  => $cgi,
	 XDAT  => $database,
	 XCKID => undef
	};

	# Check for errors
	if ($error = $cgi->cgi_error)
	{
		# Display error message
		print $cgi->header(-status=>$error); 
	
		# Signal error
		warn($error);

		# Terminate prematurely
		exit;
	}

	# Bless $self into designated class
	bless $self, $class;

	# Return self
	return $self;
}

######################
# File locking
sub LOCK_SH { 1 }
sub LOCK_EX { 2 }
sub LOCK_NB { 4 }
sub LOCK_UN { 8 }


#####################
# Cleanup
#
# This function will allow one to clear parameters in the list
# stored within this object.  Essentially it allows to remove based
# on simple keywords.  Much cleaner

sub Cleanup
{
	#####################
	# Data members
	my $self     = shift;					# Self reference
	my $key;						# Parameter key


	####################
	# Program area

	# Cycle and remove
	while ($key = shift(@_))
	{
		# Simply hiding the built in handler
		delete $self->{$key};
	}
}

#####################
# Crypt
#
# Returns the straight encoded line instead of the token like system put into place.
# This is used for scripts that need not worry about things like MORI, registration
# and such will benefit from this.

sub Crypt
{
	#####################
	# Data members
	my $self     = shift;					# Self reference


	#####################
	# Program Area

	# Return formatted string
	return "\?SEC\=" . $self->EncodeBase64(@_) . '&B64=True';
}


#####################
# Embed Information
#
# This function will embed retreived parameters for the next pass.    By making use of
# the Cypto module the use of text mangling and unmangling is no longer needed.

sub Embed
{
	#####################
	# Data members
	my $self     = shift;					# Self reference
	my $choice   = shift;					# Choice of the matter

	my $cgi      = $self->{'XCGI'};				# CGI handle
	my $database = $self->{'XDAT'};				# Database handle

	my $param;						# Parameter to embed
	my $rand;						# Random value
	my $seq;						# Sequence number
	my $string;						# Encoded parameter string
	my $user;						# User login name

	my $pname = 'XLOG';					# Parameter name
	my $puser;						# Parameter name

	my $key;						# Singular parameter key
	my $val;						# Value

        my @keys  = qw(ROOM COMM MAIN);				# Re-occuring info

	my %param;						# Additional cruft

	#####################
	# Program Area

	# Legacy random
	srand();

	# Populate hash
	%param = @_;
	

	# Safety
	$puser = (defined($param{'USER'})) ? $param{'USER'} : 'system';
	$user  = $puser;


	# Basic decisions
	# Determine parameter name
	if (defined($self->{'XCKID'}))
	{
		# Set name
		$pname = 'CKID';

		# Reset username
		$puser  = $self->{'XCKID'};
	}

	# Automatic default
	# Is this random number used
	do
	{
		# Determin random number
		$rand = int(rand() * 8000000);

		# Verify for existence
		($seq) = $database->DataGetSequence($user, $rand);

	} while (defined($seq));


	# Create embded
	# Determin style
	if ($choice < 1)
	{
		# Embeded parameter list
		$string = $cgi->hidden($pname, $puser) . ' ' . $cgi->hidden('SEC', $rand);

		# Possible additions
		foreach $key (@keys)
		{
			# Key exists add
			if (defined($param{$key}))
			{
				# Append to string
				$string .= ' ' . $cgi->hidden($key, $param{$key});

				# Delete key
				delete $param{$key};
			}
		}
	}

	# Inline parameter list
	else
	{
		# Escaped
		$val = (defined($puser)) ? $puser : $cgi->escape($user);

		# Inline parameter lists
		$string = "\?$pname\=$val\&SEC\=$rand";

		# Possible additions
		foreach $key (@keys)
		{
			# Key exists add
			if (defined($param{$key}))
			{
				# Escaped
				$val = $cgi->escape($param{$key});

				# Append to string
				$string = $string . "\&$key\=$val";

				# Delete key
				delete $param{$key};
			}
		}
	}


	# Recording medium
	# Default database method

	# Retreive encoded string
	$param = $self->EncodeString(@_, MORI=>time);
	
	# Database entrance
	$database->Quick("INSERT INTO Param
		(ParamNumber,
		 PuppeteerLogin,
		 ParamValue)
		VALUES (?,?,?)",
	 $rand,
	 $user,
	 $param
 	);

	# Return value
	return $string;
}

# Alias for Normal
sub EmbedNormal
{ return Embed(shift, 0, @_); }

# Alias for Inline
sub EmbedInline
{ return Embed(shift, 1, @_); }


#####################
# Flatten
#
# As a way to avoid the extreme annoyance of the parameter system dying out, this
# sub-routine will attempt to send all necessary parameters to be embeded.

sub Flat
{
	#####################
	# Data members
	my $self     = shift;					# Self reference
	my $choice   = shift;					# Choice in the matter

	my $key;						# Singular parameter key
	my $reg;						# Compiled regular expression

	my %param;						# Parameter hash

	#####################
	# Data members

	# Compile the regular expression checking
	# Default is A-Z otherwise look for OpenID specifics
	$reg = (!defined($choice))
	  ? qr/^[A-Z]+$/
	  : qr/^openid\.[a-z\.\_]+$/;

	#
	# Build holding hash
	foreach $key (keys(%{$self}))
	{
		# Skips
		next if ($key eq 'XCKID');
		next if ($key eq 'XCGI');
		next if ($key eq 'XDAT');
		next if ($key !~ $reg);

		# Insert
		$param{$key} = $self->{$key};
	}


	# Return flat hash
	return %param;
}

# Flatten specific for OpenID implementation
sub OpenID_Flat
{ return Flat(shift, 1); }


#####################
# Get Parameters
#
# Due to the annoyance of having to retrieve parameters from the $cgi->param
# command, I have opted to move them directly into a hash.  Passable to functions
# with little or no trouble and generally avoiding having the same value 
# re-applied to the parameter, causing problems for changes and easier calling
# via a hash.

sub GetParam
{
	#####################
	# Data members	
	my $self     = shift;					# Reference to self

	my $cgi      = $self->{'XCGI'};				# Common gateway interface
	my $database = $self->{'XDAT'};				# Database handle

	my $param;						# Specific parameter
	my $string;						# String of parameters

	my $cookie;						# Cookie based logins
	my $login;						# Login name

	my @names;						# Array of named parameters reveived
	my @vals;						# Array of pulled values


	#####################
	# Program area

	# Security measure
	$cgi->delete('CRYPT');
	$cgi->delete('MORI');

	# Initial verification
	$string = $cgi->param('SEC');


	# Method of extraction
	# Database handler
	if ((defined($string)) 
	 && ($string =~ /^\d+$/))
	{
		# Pull cookie and login variants
		$cookie = $cgi->param('CKID');
		$login  = $cgi->param('XLOG');


		# Session based cookies
		# Pull cookie
		$login = $cgi->cookie(-name=>"EREALMS\_$cookie")
		  if ((defined($cookie)) && ($cookie =~ /^\d+$/));


		# Pull complete list of attributes
		($string) = $database->DataGetParam($login, $string)
		  if (defined($login));

		# Cleanup
		$cgi->delete('XLOG');
		$cgi->delete('CKID');
	}

	# Decode and add to hash
	if ((defined($string))
	 && ($string !~ /^\d+$/))
	{
		# Base64 encoding
		if (defined($cgi->param('B64')))
	 	{
			# Create and decode
			$self->DecodeBase64($string);

			# Cleanup
			$self->Cleanup('METHOD');
		}

		# Default behaviour
		else
		{
			# Create and decode
			$self->DecodeString($string);
		}


		# Safety
		# Re-Embed
		$self->{'XCGI'}  = $cgi;
		$self->{'XDAT'}  = $database;
		$self->{'XCKID'} = $cookie;

		# Precautionary measures
		$cgi->delete('SEC');
	}

	# Get list of all parameters received
	@names = $cgi->param();	

	# Cycle though parameter list
	while ($param = shift(@names))
	{
		# Parameter pull
		# Retreive into a holding array
		@vals = $cgi->param($param);

		# Form elements are preffered
		$self->{$param} = pop(@vals);


		# Cleanup
		# Delete
		$cgi->delete($param);

		# Empty
		@vals = [];
	}
}



#####################
# Bio decoder
#
# Will take a reference to a hash, seperate it together using the implausible
# character of a form feed as the seperator.  Returns to a referenced hash.

sub Bio
{
	#####################
	# Data members
	my $self	= shift;				# Self referencing html
	my $string	= shift;				# Encoded parameter list
	my $attr        = shift;				# Hash reference

	#####################
	# Program area

	# Legacy handling
	eval
	{
		# Thaw out
		%{$attr} = %{thaw(pack('H*', $string))};
	};

	# Return true
	return 1;
}


#####################
# Cookie Decoder
#
# Will take a reference to a hash, seperate it together using the implausible
# character of a form feed as the seperator.   Is more careful in the copy
# then the regular decode.

sub Cookie
{
	#####################
	# Data members
	my $self	= shift;				# Self referencing html
	my $string	= shift;				# Encoded parameter list

	my $key;						# Temporary key holder
	my %attr;						# Cookies attributes

	#####################
	# Program area

	# Thaw out
	%attr = %{thaw(pack('H*', $string))};

	# Push over
	foreach $key (keys(%attr))
	{
		# Indirect copy to avoid damage
		$self->{$key} = $attr{$key};
	}
	
	# Return true
	return 1;
}


#####################
# Decode Parameter
#
# Will take a reference to a hash, seperate it together using the implausible
# character of a form feed as the seperator.

sub Decode
{
	#####################
	# Data members
	my $self	= shift;				# Self referencing html
	my $string	= shift;				# Encoded parameter list
	my $choice      = shift;				# Choice of the matter

	#####################
	# Program area

	# Thaw out
	# String encoding
	if ($choice eq 'string')
	{
		# Was wrapped in a string
		%{$self} = %{thaw(pack('H*', $string))};
	}

	# Base64
	elsif ($choice eq 'base64')
	{
		# Replace special characters
		$string =~ tr|\-_\t-\x0d |+/|d;

		# Re-append
		my $mod4 = length($string) % 4;
		$string .= substr('====', $mod4)
		  if ($mod4); 

		# Was encoded Base64
		%{$self} = %{thaw(decode_base64($string))};
	}

	# Straight binary
	else
	{
		# Formatted as store handled it
		%{$self} = %{thaw($string)};
	}
		
	# Return true
	return 1;
}

# Alias for bare encoding
sub DecodeBare
 {return Decode(shift, shift, 'bare'); }

# Alias for stringy handling
sub DecodeString
{ return Decode(shift, shift, 'string'); }

# Alias for Base64 handling
sub DecodeBase64
{ return Decode(shift, shift, 'base64'); }

#####################
# Encode Parameter
#
# Will take a flatened hash (key/pair), join it together using form feed as the
# seperator, encoded the entire line using the provided key.

sub Encode
{
	#####################
	# Data members
	my $self	= shift;				# Self reference
	my $choice      = shift;				# Choice of handling

	my $freeze;						# Frozen line
	my $rtn;						# Return value

	#####################
	# Program area

	# Encode string
	$freeze = nfreeze({@_});

	# Encode
	# Stringify
	return unpack('H*', $freeze)
	  if ($choice eq 'string');

	# Base64 encoding
	if ($choice eq 'base64')
	{
		# Return string
		$rtn = encode_base64($freeze, '');

		# Make safe
		$rtn =~ tr|+/=|\-_|d;


		# Return escaped
		return $rtn;

	}

	# Return string
	return $freeze;
}

# Alias for bare encoding
sub EncodeBare
 {return Encode(shift, 'bare', @_); }

# Alias for stringy handling
sub EncodeString
{ return Encode(shift, 'string', @_); }

# Alias for Base64 handling
sub EncodeBase64
{ return Encode(shift, 'base64', @_); }

1;								# Return true
